@@ -3,6 +3,7 @@ import datetime
 import traceback
 import threading
 import concurrent.futures
+import gc
 import numpy as np
 import multiprocessing as mp
 
@@ -20,6 +21,7 @@ from .singleton import Singleton
 from .subtitle import Subtitle
 from .hyperparameters import Hyperparameters
 from .exception import TerminalException
+from .exception import NoFrameRateException
 from .logger import Logger
 
 
@@ -36,6 +38,8 @@ class Predictor(Singleton):
     )  # Average 0.3 word per sec multiplies average 6 characters per word
     __MAX_HEAD_ROOM = 20000  # Maximum duration without subtitle (10 minutes)
 
+    __SEGMENT_PREDICTION_TIMEOUT = 60  # Maximum waiting time in seconds when predicting each segment
+
     def __init__(self, **kwargs):
         """Feature predictor initialiser.
 
@@ -48,7 +52,7 @@ class Predictor(Singleton):
         """
 
         self.__feature_embedder = FeatureEmbedder(**kwargs)
-        self.__lock = threading.Lock()
+        self.__lock = threading.RLock()
 
     def predict_single_pass(
             self,
@@ -74,12 +78,12 @@ class Predictor(Singleton):
             subs, audio_file_path, voice_probabilities = self.__predict(
                 video_file_path, subtitle_file_path, weights_file_path
             )
-        except Exception:
-            raise
-        else:
-            frame_rate = MediaHelper.get_frame_rate(video_file_path)
-            self.__feature_embedder.step_sample = 1 / frame_rate
-            self.__on_frame_timecodes(subs)
+            try:
+                frame_rate = MediaHelper.get_frame_rate(video_file_path)
+                self.__feature_embedder.step_sample = 1 / frame_rate
+                self.__on_frame_timecodes(subs)
+            except NoFrameRateException:
+                Predictor.__LOGGER.warn("Cannot find frame rate for %s" % video_file_path)
             return subs, audio_file_path, voice_probabilities
         finally:
             if os.path.exists(audio_file_path):
@@ -114,12 +118,13 @@ class Predictor(Singleton):
             new_subs = self.__predict_2nd_pass(
                 audio_file_path, subs, weights_file_path=weights_file_path, stretch=stretch
             )
-        except Exception:
-            raise
-        else:
-            frame_rate = MediaHelper.get_frame_rate(video_file_path)
-            self.__feature_embedder.step_sample = 1 / frame_rate
-            self.__on_frame_timecodes(new_subs)
+            try:
+                frame_rate = MediaHelper.get_frame_rate(video_file_path)
+                self.__feature_embedder.step_sample = 1 / frame_rate
+                self.__on_frame_timecodes(new_subs)
+            except NoFrameRateException:
+                Predictor.__LOGGER.warn("Cannot find frame rate for %s" % video_file_path)
+            Predictor.__LOGGER.debug("Aligned segments generated")
             return new_subs, subs, voice_probabilities
         finally:
             if os.path.exists(audio_file_path):
@@ -150,22 +155,23 @@ class Predictor(Singleton):
                     [np.zeros(voice_probabilities.shape[1])] * (-head_room * 5),
                 ]
             )
+            result = log_loss(
+                subtitle_mask, local_vp[: len(subtitle_mask)], labels=[0, 1]
+            )
         else:
-            local_vp = voice_probabilities
-        result = log_loss(
-            subtitle_mask, local_vp[: len(subtitle_mask)], labels=[0, 1]
-        )
+            result = log_loss(
+                subtitle_mask, voice_probabilities[: len(subtitle_mask)], labels=[0, 1]
+            )
+
         Predictor.__LOGGER.debug("Log loss: {}".format(result))
         return result
 
     def get_min_log_loss_and_index(self, voice_probabilities, subs):
         """Returns the minimum loss value and its shift position
         after going through all possible shifts.
-
             Arguments:
                 voice_probabilities {list} -- A list of probabilities of audio chunks being speech.
                 subs {list} -- A list of subtitle segments.
-
             Returns:
                 tuple -- The minimum loss value and its position.
         """
@@ -211,9 +217,17 @@ class Predictor(Singleton):
             )
         if log_losses:
             min_log_loss = min(log_losses)
-            return min_log_loss, log_losses.index(min_log_loss)
+            min_log_loss_idx = log_losses.index(min_log_loss)
         else:
-            return None, 0
+            min_log_loss = None
+            min_log_loss_idx = 0
+
+        del local_vp
+        del log_losses
+        gc.collect()
+
+        return min_log_loss, min_log_loss_idx
+
 
     def __predict(
             self,
@@ -242,6 +256,7 @@ class Predictor(Singleton):
             tuple -- The shifted subtitles, the audio file path and the voice probabilities of the original audio.
         """
 
+        thread_name = threading.current_thread().name
         result = {}
         pred_start = datetime.datetime.now()
         if audio_file_path is not None:
@@ -252,8 +267,8 @@ class Predictor(Singleton):
                 video_file_path, True, 16000
             )
             Predictor.__LOGGER.debug(
-                "Audio extracted after {}".format(
-                    str(datetime.datetime.now() - t)
+                "[{}] Audio extracted after {}".format(
+                    thread_name, str(datetime.datetime.now() - t)
                 )
             )
             result["video_file_path"] = video_file_path
@@ -269,15 +284,15 @@ class Predictor(Singleton):
             if os.path.exists(audio_file_path):
                 os.remove(audio_file_path)
             raise TerminalException("Error: No subtitles passed in")
-
-        try:
-            train_data, labels = self.__feature_embedder.extract_data_and_label_from_audio(
-                audio_file_path, None, subtitles=subs
-            )
-        except TerminalException:
-            if os.path.exists(audio_file_path):
-                os.remove(audio_file_path)
-            raise
+        with self.__lock:
+            try:
+                train_data, labels = self.__feature_embedder.extract_data_and_label_from_audio(
+                    audio_file_path, None, subtitles=subs
+                )
+            except TerminalException:
+                if os.path.exists(audio_file_path):
+                    os.remove(audio_file_path)
+                raise
 
         train_data = np.array([np.rot90(val) for val in train_data])
         train_data = train_data - np.mean(train_data, axis=0)
@@ -286,16 +301,21 @@ class Predictor(Singleton):
 
         # Load neural network
         input_shape = (train_data.shape[1], train_data.shape[2])
-        Predictor.__LOGGER.debug("input shape: {}".format(input_shape))
+        Predictor.__LOGGER.debug("[{}] input shape: {}".format(thread_name, input_shape))
 
         # Network class is not thread safe so a new graph is created for each thread
-        self.__lock.acquire()
-        try:
-            Predictor.__LOGGER.debug("Start predicting...")
-            pred_start = datetime.datetime.now()
-            voice_probabilities = self.__network.get_predictions(train_data, weights_file_path)
-        finally:
-            self.__lock.release()
+        pred_start = datetime.datetime.now()
+        with self.__lock:
+            try:
+                Predictor.__LOGGER.debug("[{}] Start predicting...".format(thread_name))
+                voice_probabilities = self.__network.get_predictions(train_data, weights_file_path)
+            except Exception as e:
+                Predictor.__LOGGER.error("[{}] Prediction failed: {}\n{}".format(thread_name, str(e), traceback.format_stack()))
+                raise TerminalException("Prediction failed") from e
+            finally:
+                del train_data
+                del labels
+                gc.collect()
 
         if len(voice_probabilities) <= 0:
             if os.path.exists(audio_file_path):
@@ -310,20 +330,20 @@ class Predictor(Singleton):
         # Predictor.__LOGGER.debug("predictions: {}".format(voice_probabilities))
 
         original_start = FeatureEmbedder.time_to_sec(subs[0].start)
-        Predictor.__LOGGER.debug("Start {}".format(original_start))
         shifted_subs = deepcopy(subs)
         subs.shift(seconds=-original_start)
 
-        Predictor.__LOGGER.info("Aligning subtitle with video...")
+        Predictor.__LOGGER.info("[{}] Aligning subtitle with video...".format(thread_name))
 
-        min_log_loss, min_log_loss_pos = self.get_min_log_loss_and_index(
-            voice_probabilities, subs
-        )
+        with self.__lock:
+            min_log_loss, min_log_loss_pos = self.get_min_log_loss_and_index(
+                voice_probabilities, subs
+            )
 
         pos_to_delay = min_log_loss_pos
         result["loss"] = min_log_loss
 
-        Predictor.__LOGGER.info("Subtitle aligned")
+        Predictor.__LOGGER.info("[{}] Subtitle aligned".format(thread_name))
 
         if subtitle_file_path is not None:  # for the first pass
             seconds_to_shift = (
@@ -349,11 +369,11 @@ class Predictor(Singleton):
         result["original_start"] = original_start
         total_elapsed_time = str(datetime.datetime.now() - pred_start)
         result["time_sync"] = total_elapsed_time
-        Predictor.__LOGGER.debug("Statistics: {}".format(result))
+        Predictor.__LOGGER.debug("[{}] Statistics: {}".format(thread_name, result))
 
-        Predictor.__LOGGER.debug("Total Time: {}".format(total_elapsed_time))
+        Predictor.__LOGGER.debug("[{}] Total Time: {}".format(thread_name, total_elapsed_time))
         Predictor.__LOGGER.debug(
-            "Seconds to shift: {}".format(seconds_to_shift)
+            "[{}] Seconds to shift: {}".format(thread_name, seconds_to_shift)
         )
 
         # For each subtitle chunk, its end time should not be later than the end time of the audio segment
@@ -361,11 +381,12 @@ class Predictor(Singleton):
             shifted_subs.shift(seconds=seconds_to_shift)
         elif max_shift_secs is not None and seconds_to_shift > max_shift_secs:
             Predictor.__LOGGER.warning(
-                "Maximum {} seconds shift has reached".format(max_shift_secs)
+                "[{}] Maximum {} seconds shift has reached".format(thread_name, max_shift_secs)
             )
             shifted_subs.shift(seconds=max_shift_secs)
         else:
             shifted_subs.shift(seconds=seconds_to_shift)
+        Predictor.__LOGGER.debug("[{}] Subtitle shifted".format(thread_name))
         return shifted_subs, audio_file_path, voice_probabilities
 
     def __predict_2nd_pass(self, audio_file_path, subs, weights_file_path, stretch):
@@ -378,9 +399,7 @@ class Predictor(Singleton):
             stretch {bool} -- True to stretch the subtitle segments (default: {False})
         """
 
-        segment_starts, segment_ends, subs = MediaHelper.get_audio_segment_starts_and_ends(
-            subs
-        )
+        segment_starts, segment_ends, subs = MediaHelper.get_audio_segment_starts_and_ends(subs)
         subs_copy = deepcopy(subs)
 
         for index, sub in enumerate(subs):
@@ -400,7 +419,10 @@ class Predictor(Singleton):
         subs_list = []
 
         max_workers = int(os.getenv("MAX_WORKERS", mp.cpu_count() / 2))
-        with concurrent.futures.ThreadPoolExecutor(
+        Predictor.__LOGGER.debug("Number of workers: {}".format(max_workers))
+
+        with _ThreadPoolExecutorLocal(
+                queue_size=10,
                 max_workers=max_workers
         ) as executor:
             futures = [
@@ -418,23 +440,30 @@ class Predictor(Singleton):
                 )
                 for i in range(len(segment_starts))
             ]
-            concurrent.futures.wait(futures)
-            for future in futures:
+            for i, future in enumerate(futures):
                 try:
-                    new_subs = future.result()
+                    new_subs = future.result(timeout=Predictor.__SEGMENT_PREDICTION_TIMEOUT)
+                except concurrent.futures.TimeoutError as e:
+                    self.__cancel_futures(futures[i:], Predictor.__SEGMENT_PREDICTION_TIMEOUT)
+                    message = "Segment alignment timed out after {} seconds".format(Predictor.__SEGMENT_PREDICTION_TIMEOUT)
+                    Predictor.__LOGGER.error(message)
+                    raise TerminalException(message) from e
                 except Exception as e:
-                    Predictor.__LOGGER.error(
-                        "Exception: {}\n{}".format(
-                            str(e), traceback.format_stack()
-                        )
-                    )
+                    self.__cancel_futures(futures[i:], Predictor.__SEGMENT_PREDICTION_TIMEOUT)
+                    message = "Exception on segment alignment: {}\n{}".format(str(e), traceback.format_stack())
+                    Predictor.__LOGGER.error(message)
+                    raise TerminalException(message) from e
+                except KeyboardInterrupt:
+                    self.__cancel_futures(futures[i:], Predictor.__SEGMENT_PREDICTION_TIMEOUT)
                     raise
                 else:
+                    Predictor.__LOGGER.debug("Segment aligned")
                     subs_list.append(new_subs)
 
         subs_list = [
             sub_item for subs in subs_list for sub_item in subs
         ]  # flatten the subs_list
+        Predictor.__LOGGER.debug("All segments aligned")
         return subs_list
 
     def __predict_in_multithreads(
@@ -448,6 +477,7 @@ class Predictor(Singleton):
             subs_copy,
             stretch,
     ):
+        thread_name = threading.current_thread().name
         segment_path = ""
         try:
             if segment_index == (len(segment_starts) - 1):
@@ -475,7 +505,7 @@ class Predictor(Singleton):
                     subs[segment_index - 1][len(subs[segment_index - 1]) - 1].end
                 )
 
-            subs_new, _, _ = self.__predict(
+            subs_new, _, voice_probabilities = self.__predict(
                 video_file_path=None,
                 subtitle_file_path=None,
                 weights_file_path=weights_file_path,
@@ -484,15 +514,30 @@ class Predictor(Singleton):
                 max_shift_secs=max_shift_secs,
                 previous_gap=previous_gap,
             )
+            del voice_probabilities
+            gc.collect()
 
             if stretch:
                 subs_new = self.__adjust_durations(subs_new, audio_file_path)
+                Predictor.__LOGGER.info("[{}] Segment {} stretched".format(thread_name, segment_index))
             return subs_new
+        except Exception as e:
+            Predictor.__LOGGER.error(
+                "[{}] Alignment failed for segment {}: {}\n{}".format(
+                    thread_name, segment_index, str(e), traceback.format_stack()
+                )
+            )
+            return subs[segment_index]
         finally:
             # Housekeep intermediate files
             if os.path.exists(segment_path):
                 os.remove(segment_path)
 
+    def __cancel_futures(self, futures, timeout):
+        for future in futures:
+            future.cancel()
+        concurrent.futures.wait(futures, timeout=timeout)
+
     def __get_subtitle_mask(self, subs):
         pos = self.__feature_embedder.time_to_pos(subs[len(subs) - 1].end) - 1
         subtitle_mask = np.zeros(pos if pos > 0 else 0)
@@ -548,35 +593,7 @@ class Predictor(Singleton):
             hyperparams = Hyperparameters.from_file(hyperparams_path)
             self.__network = Network.get_from_model(model_path, hyperparams)
 
-    @staticmethod
-    def __get_weights_path(weights_dir):
-        weights_dir = os.path.join(os.path.dirname(__file__), weights_dir)
-        files = os.listdir(weights_dir)
-        weights_files = [
-            file
-            for file in files
-            if file.startswith("weights")
-        ]
-
-        if not weights_files:
-            raise TerminalException(
-                "Cannot find weights files at {}".format(weights_dir)
-            )
-
-        Predictor.__LOGGER.debug("weights files: {}".format(weights_files))
-
-        # Get the first file from the file lists
-        weights_path = "{}/{}".format(weights_dir, weights_files[0])
-
-        return os.path.join(os.path.dirname(__file__), weights_path)
-
-    @staticmethod
-    def __normalise_seconds_to_shift(seconds_to_shift, step_sample):
-        # Make sure each cue starts right on the beginning of a frame
-        return round(seconds_to_shift / step_sample) * step_sample
-
-    @staticmethod
-    def __adjust_durations(subs, audio_file_path):
+    def __adjust_durations(self, subs, audio_file_path):
 
         # Initialise a DTW alignment task
         task_config_string = (
@@ -605,15 +622,21 @@ class Predictor(Singleton):
             task.text_file_path_absolute = text_file_path
             task.sync_map_file_path_absolute = "{}.srt".format(root)
 
-            # Execute the task
-            ExecuteTask(
-                task=task,
-                rconf=RuntimeConfiguration(config_string=runtime_config_string),
-                logger=AeneasLogger(tee=Logger.VERBOSE),
-            ).execute()
+            tee = False
+            if Logger.VERBOSE:
+                tee = True
+            if Logger.QUIET:
+                tee = False
+            with self.__lock:
+                # Execute the task
+                ExecuteTask(
+                    task=task,
+                    rconf=RuntimeConfiguration(config_string=runtime_config_string),
+                    logger=AeneasLogger(tee=tee),
+                ).execute()
 
-            # Output new subtitle segment to a file
-            task.output_sync_map_file()
+                # Output new subtitle segment to a file
+                task.output_sync_map_file()
 
             # Load the above subtitle segment
             adjusted_subs = Subtitle.load(
@@ -640,3 +663,214 @@ class Predictor(Singleton):
                 os.remove(task.text_file_path_absolute)
             if task.sync_map_file_path_absolute is not None and os.path.exists(task.sync_map_file_path_absolute):
                 os.remove(task.sync_map_file_path_absolute)
+
+    @staticmethod
+    def __get_weights_path(weights_dir):
+        weights_dir = os.path.join(os.path.dirname(__file__), weights_dir)
+        files = os.listdir(weights_dir)
+        weights_files = [
+            file
+            for file in files
+            if file.startswith("weights")
+        ]
+
+        if not weights_files:
+            raise TerminalException(
+                "Cannot find weights files at {}".format(weights_dir)
+            )
+
+        Predictor.__LOGGER.debug("weights files: {}".format(weights_files))
+
+        # Get the first file from the file lists
+        weights_path = "{}/{}".format(weights_dir, weights_files[0])
+
+        return os.path.join(os.path.dirname(__file__), weights_path)
+
+    @staticmethod
+    def __normalise_seconds_to_shift(seconds_to_shift, step_sample):
+        # Make sure each cue starts right on the beginning of a frame
+        return round(seconds_to_shift / step_sample) * step_sample
+
+    def __predict(
+            self,
+            video_file_path,
+            subtitle_file_path,
+            weights_file_path,
+            audio_file_path=None,
+            subtitles=None,
+            max_shift_secs=None,
+            previous_gap=None,
+    ):
+        """Shift out-of-sync subtitle cues by sending the audio track of an video to the trained network.
+
+        Arguments:
+            video_file_path {string} -- The file path of the original video.
+            subtitle_file_path {string} -- The file path of the out-of-sync subtitles.
+            weights_file_path {string} -- The file path of the weights file.
+
+        Keyword Arguments:
+            audio_file_path {string} -- The file path of the original audio (default: {None}).
+            subtitles {list} -- The list of SubRip files (default: {None}).
+            max_shift_secs {float} -- The maximum seconds by which subtitle cues can be shifted (default: {None}).
+            previous_gap {float} -- The duration betwee the start time of the audio segment and the start time of the subtitle segment.
+
+        Returns:
+            tuple -- The shifted subtitles, the audio file path and the voice probabilities of the original audio.
+        """
+
+        thread_name = threading.current_thread().name
+        result = {}
+        pred_start = datetime.datetime.now()
+        if audio_file_path is not None:
+            result["audio_file_path"] = audio_file_path
+        elif video_file_path is not None:
+            t = datetime.datetime.now()
+            audio_file_path = MediaHelper.extract_audio(
+                video_file_path, True, 16000
+            )
+            Predictor.__LOGGER.debug(
+                "[{}] Audio extracted after {}".format(
+                    thread_name, str(datetime.datetime.now() - t)
+                )
+            )
+            result["video_file_path"] = video_file_path
+        else:
+            raise TerminalException("Neither audio nor video is passed in")
+
+        if subtitle_file_path is not None:
+            subs = Subtitle.load(subtitle_file_path).subs
+            result["subtitle_file_path"] = subtitle_file_path
+        elif subtitles is not None:
+            subs = subtitles
+        else:
+            if os.path.exists(audio_file_path):
+                os.remove(audio_file_path)
+            raise TerminalException("Error: No subtitles passed in")
+        with self.__lock:
+            try:
+                train_data, labels = self.__feature_embedder.extract_data_and_label_from_audio(
+                    audio_file_path, None, subtitles=subs
+                )
+            except TerminalException:
+                if os.path.exists(audio_file_path):
+                    os.remove(audio_file_path)
+                raise
+
+        train_data = np.array([np.rot90(val) for val in train_data])
+        train_data = train_data - np.mean(train_data, axis=0)
+        result["time_load_dataset"] = str(datetime.datetime.now() - pred_start)
+        result["X_shape"] = train_data.shape
+
+        # Load neural network
+        input_shape = (train_data.shape[1], train_data.shape[2])
+        Predictor.__LOGGER.debug("[{}] input shape: {}".format(thread_name, input_shape))
+
+        # Network class is not thread safe so a new graph is created for each thread
+        pred_start = datetime.datetime.now()
+        with self.__lock:
+            try:
+                Predictor.__LOGGER.debug("[{}] Start predicting...".format(thread_name))
+                voice_probabilities = self.__network.get_predictions(train_data, weights_file_path)
+            except Exception as e:
+                Predictor.__LOGGER.error("[{}] Prediction failed: {}\n{}".format(thread_name, str(e), traceback.format_stack()))
+                raise TerminalException("Prediction failed") from e
+            finally:
+                del train_data
+                del labels
+                gc.collect()
+
+        if len(voice_probabilities) <= 0:
+            if os.path.exists(audio_file_path):
+                os.remove(audio_file_path)
+            raise TerminalException(
+                "Error: Audio is too short and no voice was detected"
+            )
+
+        result["time_predictions"] = str(datetime.datetime.now() - pred_start)
+
+        # for p in voice_probabilities: Predictor.__LOGGER.debug("{}, ".format(p))
+        # Predictor.__LOGGER.debug("predictions: {}".format(voice_probabilities))
+
+        original_start = FeatureEmbedder.time_to_sec(subs[0].start)
+        shifted_subs = deepcopy(subs)
+        subs.shift(seconds=-original_start)
+
+        Predictor.__LOGGER.info("[{}] Aligning subtitle with video...".format(thread_name))
+
+        with self.__lock:
+            min_log_loss, min_log_loss_pos = self.get_min_log_loss_and_index(
+                voice_probabilities, subs
+            )
+
+        pos_to_delay = min_log_loss_pos
+        result["loss"] = min_log_loss
+
+        Predictor.__LOGGER.info("[{}] Subtitle aligned".format(thread_name))
+
+        if subtitle_file_path is not None:  # for the first pass
+            seconds_to_shift = (
+                self.__feature_embedder.pos_to_sec(pos_to_delay) - original_start
+            )
+        elif subtitles is not None:  # for each in second pass
+            seconds_to_shift = self.__feature_embedder.pos_to_sec(pos_to_delay) - previous_gap
+        else:
+            if os.path.exists(audio_file_path):
+                os.remove(audio_file_path)
+            raise ValueError("Error: No subtitles passed in")
+
+        if abs(seconds_to_shift) > Predictor.__MAX_SHIFT_IN_SECS:
+            if os.path.exists(audio_file_path):
+                os.remove(audio_file_path)
+            raise TerminalException(
+                "Average shift duration ({} secs) have been reached".format(
+                    Predictor.__MAX_SHIFT_IN_SECS
+                )
+            )
+
+        result["seconds_to_shift"] = seconds_to_shift
+        result["original_start"] = original_start
+        total_elapsed_time = str(datetime.datetime.now() - pred_start)
+        result["time_sync"] = total_elapsed_time
+        Predictor.__LOGGER.debug("[{}] Statistics: {}".format(thread_name, result))
+
+        Predictor.__LOGGER.debug("[{}] Total Time: {}".format(thread_name, total_elapsed_time))
+        Predictor.__LOGGER.debug(
+            "[{}] Seconds to shift: {}".format(thread_name, seconds_to_shift)
+        )
+
+        # For each subtitle chunk, its end time should not be later than the end time of the audio segment
+        if max_shift_secs is not None and seconds_to_shift <= max_shift_secs:
+            shifted_subs.shift(seconds=seconds_to_shift)
+        elif max_shift_secs is not None and seconds_to_shift > max_shift_secs:
+            Predictor.__LOGGER.warning(
+                "[{}] Maximum {} seconds shift has reached".format(thread_name, max_shift_secs)
+            )
+            shifted_subs.shift(seconds=max_shift_secs)
+        else:
+            shifted_subs.shift(seconds=seconds_to_shift)
+        Predictor.__LOGGER.debug("[{}] Subtitle shifted".format(thread_name))
+        return shifted_subs, audio_file_path, voice_probabilities
+
+
+class _ThreadPoolExecutorLocal:
+
+    def __init__(self, queue_size, max_workers):
+        self.executor = concurrent.futures.ThreadPoolExecutor(max_workers=max_workers)
+        self.semaphore = threading.BoundedSemaphore(queue_size + max_workers)
+
+    def __enter__(self):
+        return self
+
+    def __exit__(self, exc_type, exc_val, exc_tb):
+        self.executor.shutdown(True)
+
+    def submit(self, fn, *args, **kwargs):
+        self.semaphore.acquire()
+        try:
+            future = self.executor.submit(fn, *args, **kwargs)
+        except Exception:
+            self.semaphore.release()
+            raise
+        else:
+            future.add_done_callback(lambda x: self.semaphore.release())
+            return future
